/*
  gmap.cpp
  Copyright (C) 2008 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include <qwidget.h>
#include <qcursor.h>
#include <qpainter.h>
#include <stdio.h>
#include <stdlib.h>
#include "qgmap.h"
#include "gmap.h"

#include <qdialog.h>
#include <qdatetime.h>

#include "pos.h"
#include "inputdialog.h"

GMap::GMap( QWidget *parent, const char *name, WFlags fl)
	: QWidget( parent, name, fl), img_cache(31,31),
	  info_label(this),
	  marker(this, &center),
	  help_label(this)
{
//#ifndef QWS
	resize(640,640);
//#endif
	setCaption(QString::fromUtf8("qgmap"));

	is_map_click = FALSE;
	is_scale_click = FALSE;

	info_label.hide();
	info_label.setFrameStyle(QFrame::Box | QFrame::Plain);
	info_label.setFont(QFont("lcfont", 12));

	help_label.hide();
	help_label.setFrameStyle(QFrame::Box | QFrame::Plain);
	help_label.setFont(QFont("lcfont", 16));
	help_label.setMargin(2);
	help_label.setText(QString::fromUtf8(
		"Help\n"
		" カーソル : 上下左右スクロール (Shift+カーソルで大幅移動)\n"
		"   I,O    : ズームイン, アウト\n"
		"   0-9    : 保存位置読み出し\n"
		"Shift+0-9 : 位置保存 (0 は初期位置)\n"
		"    J     : 緯度・経度指定ジャンプ\n"
		"    M     : マーカ表示ON/OFF\n"
		"            (ツールチップ表示中に 0-9 キーで個別マーカON/OFF)\n"
		"  H or ?  : ヘルプ表示"
						   ));
	help_label.adjustSize();
	is_help = false;

	info_timer_id = -1;

	popup_menu.insertItem(tr("Map Path"), this, SLOT(menu_activated(int)));

	map_path = conf.readEntry("MapPath", DEFAULT_MAP_PATH);
//	printf("MapPath = %s\n", map_path.ascii());

	menu_pic.load(MENUIMG_FILE);

	// 8MB Cache
	// 256x256x32bpp = 256kB/Picture
	// 8MB/256kB = 32 Picture
	cache.setCacheLimit(8192);

	img_cache.setAutoDelete(TRUE);

	loadPosition(0);

	scale = new Scale(&scr, &center);
	calcScale();
}

void GMap::updateCenter(bool force)
{
	int widget_top, widget_left;
	int new_ltop, new_lleft;

	// ウィジェットの左上部の地図座標
	widget_top = center.getY()-this->height()/2;
	widget_left = center.getX()-this->width()/2;
//	printf("widget_left=%d, widget_top=%d\n", widget_left, widget_top);

	// 読み込み画像の左上の地図座標
	new_ltop = widget_top & ~(IMAGE_SIZE-1);
	new_lleft = widget_left & ~(IMAGE_SIZE-1);
//	printf("load_left=%d, load_top=%d\n", load_left, load_top);
//	printf("new_lleft=%d, new_ltop=%d force=%d\n", new_lleft, new_ltop, force);

	Pos pos(new_lleft, new_ltop, center.getZoom());

//	QTime t;
	// 強制読み込み時、スクロールしたときロード時
	if(load_left != new_lleft || load_top != new_ltop || force){

//		t.start();
		//全画像ロード
		loadMap(pos);
		//内部で load_left, load_top を更新
//		printf("forceLoadMap() %dms.\n", t.elapsed());
	}

	// 再描画指示(消去しない)
	repaint(FALSE);
}

void GMap::paintEvent( QPaintEvent *e )
{
	int x,y;
	int dx, dy;

	/*
	  if ( !e->rect().intersects( cannonRect() ) )
	  return;

	  QRect cr = cannonRect();
	  QPixmap pix( cr.size() );
	*/

	dx = (center.getX() - this->width()/2) - load_left;
	dy = (center.getY() - this->height()/2) - load_top;

#if 0
	printf("GMap::paintEvent load_left=%d, load_top=%d\n", load_left, load_top);
	printf("GMap::paintEvent widget_x=%d, widget_y=%d\n",
		   (center.getX() - this->width()/2), (center.getY() - this->height()/2));
	printf("GMap::paintEvent width=%d, height=%d\n", this->width(), this->height());
	printf("GMap::paintEvent dx=%d, dy=%d\n", dx, dy);
#endif

	QPixmap pixmap(e->rect().size());
	QPainter p( &scr );
	p.setPen(Qt::black);

	for(y = 0; y < y_num; y++){
		for(x = 0; x < x_num; x++){
			p.drawPixmap( x*IMAGE_SIZE-dx, y*IMAGE_SIZE-dy, pix[y*x_num+x]->getPix(),
						  0, 0, IMAGE_SIZE, IMAGE_SIZE);
//			p.drawRect( x*IMAGE_SIZE-dx, y*IMAGE_SIZE-dy, IMAGE_SIZE+1, IMAGE_SIZE+1);
		}
	}

	p.drawPixmap( 10, 10, menu_pic);
	p.end();

	scale->paintEvent( e );
	drawScale(&scr);

	marker.paint(&scr);

	p.begin( this );
	p.drawPixmap( 0, 0, scr );

}

void GMap::wheelEvent( QWheelEvent *e )
{
	zoomup(e->delta()<0?1:-1);
}

void GMap::mouseMoveEvent( QMouseEvent *e )
{
	if(is_map_click){
		QPoint pos = click_pos - e->pos();

		center = center.offset(pos.x(), pos.y());
		center.normalize();
		updateCenter(FALSE);

		click_pos = e->pos();
	}
	if(is_scale_click){
		if(scale->moveEvent(e)){
			calcScale();
			updateCenter(TRUE);
		}
	}
}

void GMap::mousePressEvent( QMouseEvent *e )
{
	QRect menu_rect(10,10,16,16);

	if(is_help){
		hideHelp();
		return;
	}

	if(e->button() != LeftButton)
		return;

	if(!is_map_click && !is_scale_click){
		if(menu_rect.contains(e->pos())){
//			printf("click menu\n");
			popup_menu.popup(QCursor::pos());
			return;
		}else if(scale->rect().contains(e->pos())){
			if(scale->pressEvent(e)){
				calcScale();
				updateCenter(TRUE);
			}
			is_scale_click = TRUE;
			return;
		}
	}

	is_map_click = TRUE;
	click_pos = e->pos();
}

void GMap::mouseReleaseEvent( QMouseEvent *e )
{
	if(e->button() != LeftButton)
		return;

	if(is_scale_click){
		if(scale->releaseEvent(e)){
			calcScale();
			updateCenter(TRUE);
		}
		is_scale_click = FALSE;
		return;
	}

	is_map_click = FALSE;
	calcScale();
	repaint(FALSE);
}

void GMap::keyPressEvent ( QKeyEvent * e )
{
	int shift = (e->state() & ShiftButton)?SHIFT_MUL:1;

	if(is_help){
		hideHelp();
		return;
	}

	/* marker hook */
	if(marker.keyPressEvent(e))
		return;

	switch(e->key()){
	case Key_Left:
		scroll(-KEY_SCROLL*shift, 0);
		break;
	case Key_Right:
		scroll(KEY_SCROLL*shift, 0);
		break;
	case Key_Up:
		scroll(0, -KEY_SCROLL*shift);
		break;
	case Key_Down:
		scroll(0, KEY_SCROLL*shift);
		break;
	case Key_I:
		zoomup(-1);
		break;
	case Key_O:
		zoomup(1);
		break;
	case Key_J:
		jumpLocation();
		break;

	case Key_H:
		dispHelp();
		break;
	case Key_Period:
		if(ShiftButton)
			dispHelp();
		break;

	case Key_0:
	case Key_1:
	case Key_2:
	case Key_3:
	case Key_4:
	case Key_5:
	case Key_6:
	case Key_7:
	case Key_8:
	case Key_9:
		if(e->state() & ShiftButton) {
			savePosition(e->key() - Key_0);
			repaint(FALSE);
		} else {
			loadPosition(e->key() - Key_0);
			updateCenter(TRUE);
		}
		break;

#ifndef QWS	// for PC
	case Key_Exclam:
		savePosition(1);
		break;
	case Key_QuoteDbl:
		savePosition(2);
		break;
	case Key_NumberSign:
		savePosition(3);
		break;
	case Key_Dollar:
		savePosition(4);
		break;
	case Key_Percent:
		savePosition(5);
		break;
	case Key_Ampersand:
		savePosition(6);
		break;
	case Key_Apostrophe:
		savePosition(7);
		break;
	case Key_ParenLeft:
		savePosition(8);
		break;
	case Key_ParenRight:
		savePosition(9);
		break;
	case Key_AsciiTilde:
		savePosition(0);
		break;
#endif // for PC
	}
}

void GMap::keyReleaseEvent ( QKeyEvent * e )
{
	calcScale();
}

void GMap::timerEvent( QTimerEvent *e )
{
	int id;

	/* marker hook */
	if(marker.timerEvent(e))
		return;

	id = e->timerId();
	printf("timer id=%d\n", e->timerId());
	if(id == info_timer_id){
		info_label.hide();
		killTimer(id);
		info_timer_id=-1;
	}
}

/* 全画像の読み込み */
void GMap::loadMap(Pos loadpos)
{
	int x, y;

	for(y = 0; y < y_num; y++){
		for(x = 0; x < x_num; x++){
			pix[y*x_num+x]->loadPixmap(loadpos.offset(x*IMAGE_SIZE,y*IMAGE_SIZE));
		}
	}
	load_left = loadpos.getX();
	load_top = loadpos.getY();
}

void GMap::scroll(int x, int y)
{
	center = center.offset(x, y);
	updateCenter(FALSE);
}

void GMap::zoomup(int z)
{
	center.zoomup(z);
	calcScale();
	updateCenter(TRUE);
}

void GMap::menu_activated(int id)
{
//	printf("menu %d\n",id);

	InputDialog dlg(this, tr("Map Path"), TRUE);

	dlg.setText(conf.readEntry("MapPath", DEFAULT_MAP_PATH));

	int ret = dlg.exec();
//	printf("ret:%d text=%s\n", ret, dlg.getText().ascii());
	if(ret == QDialog::Accepted){
		map_path = dlg.getText();
		conf.writeEntry("MapPath", map_path);
//		printf("map_path=%s\n", map_path.ascii());
		updateCenter(TRUE);
	}
}

void GMap::calcScale()
{
	scale_size = center.getScaleSize(SCALE_MAX_SIZE, &scale_mul);
}

void GMap::drawScale(QPixmap *scr)
{
	QPainter p(scr);

	int xoff = 10, yoff = scr->height() - 24;

	p.setPen(Qt::white);
	p.drawRect(xoff, yoff+4, 4,14);
	p.drawRect(xoff+scale_size, yoff+4, 4,14);
	p.drawRect(xoff, yoff+4+10, scale_size+4,4);

	p.setPen(Qt::black);
	p.drawRect(xoff+1, yoff+4+1, 2,10);
	p.drawRect(xoff+1+scale_size, yoff+4+1, 2,10);
	p.drawRect(xoff+1, yoff+4+11, scale_size+2,2);

	QFont font = QFont::defaultFont();
	font.setPixelSize(SCALE_FONT_SIZE);
	p.setFont(font);
	QString str;
	if(scale_mul>1000)
		str.sprintf("%dkm",scale_mul/1000);
	else
		str.sprintf("%dm",scale_mul);

	p.drawText(xoff+10, yoff+10, str);
}

void GMap::jumpLocation()
{
	InputDialog dlg(this, tr("Jump Location (Lat,Lon)"), TRUE);

	int ret = dlg.exec();
	if(ret == QDialog::Accepted){
		if(center.setLatLon(dlg.getText())){
			updateCenter(TRUE);
		}
	}
}

void GMap::savePosition(int i)
{
	QString key;
	QString info_str;

	key = tr("Pos%1").arg(i);
	conf.writeEntry(key, center.serialize());

	info_str = QString::fromUtf8("Pos%1に保存").arg(i);
	dispInfo(info_str);
}

void GMap::loadPosition(int i)
{
	QString key, str;
	Pos tmp;

	key.sprintf("Pos%d", i);
	str = conf.readEntry(key, tmp.serialize().ascii());

	center.deserialize(str);
}

Pos GMap::getPosition(int i)
{
	QString key, str;
	Pos tmp;

	key.sprintf("Pos%d", i);
	str = conf.readEntry(key, tmp.serialize().ascii());

	tmp.deserialize(str);
	return tmp;
}

void GMap::dispInfo(QString str)
{
	if(info_timer_id > 0){
		killTimer(info_timer_id);
		info_timer_id=-1;
	}
	info_label.setText(str);
	info_label.adjustSize();
//	printf("label=%dx%d\n",info_label.width(), info_label.height());
	info_label.setGeometry(30, 10, info_label.width(), info_label.height());
	info_label.show();
	info_timer_id = startTimer(3000);
}

void GMap::dispHelp()
{
	int x, y;

	x = (width()-help_label.width())/2;
	y = (height()-help_label.height())/2;
	help_label.setGeometry(x, y,
						   help_label.width(), help_label.height());
	help_label.show();
	is_help = true;
}

void GMap::hideHelp()
{
	help_label.hide();
	is_help = false;
}

void GMap::showEvent( QShowEvent *e )
{
	int i;

	scr.resize(this->width(), this->height());

	x_num = (this->width()/IMAGE_SIZE)+2;
	y_num = (this->height()/IMAGE_SIZE)+2;

	pix.resize(x_num*y_num);

	for(i=0;i<x_num*y_num;i++){
		pix[i] = new Map(&cache, &img_cache, &map_path);
	}

	updateCenter(FALSE);
}

