/*
  marker.h
  Copyright (C) 2008 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <qlabel.h>
#include <qevent.h>
#include <qpen.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qfont.h>
#include "gmap.h"
#include "marker.h"

#include <math.h>

Marker::Marker(QWidget *parent, const Pos *center)
{
	QString path;
	int i;

	this->parent = parent;
	this->center = center;

	base.load(MARKER_FILE);

	for(i=0;i<9;i++){
		if(i==4) continue;

		path.sprintf(APP_PATH"/arrow%d.png", i);
		arrow[i].load(path);
	}

	marker_on = false;
	marker_bit = 0;
	marker_timer_id = -1;
}

Marker::~Marker()
{
}

void Marker::paint(QPixmap *scr)
{
	int i;
	if(!marker_on)
		return;

	for(i=0;i<10;i++){
		if(IS_MARKER_BIT(i)){
			drawMarker(scr, ((GMap *)parent)->getPosition(i), i);
		}
	}
}

void Marker::drawMarker(QPixmap *scr, Pos pos, int index)
{
	int sx,sy;

	pos.zoomTo(center->getZoom());

	sx = pos.getX()-center->getX()+scr->width()/2;
	sy = pos.getY()-center->getY()+scr->height()/2;

	if(isClip(scr, sx, sy)){
		drawClipMarker(scr, sx, sy, index);
		return;
	}

	drawBase(scr, sx, sy-8-5, index);
	QPainter p(scr);
	p.drawPixmap( sx-3, sy-7, arrow[7]);
}

void Marker::drawBase(QPixmap *scr, int cx, int cy, int index)
{
	QPainter p(scr);
	QFont font("helvetica", 12, QFont::Bold);
	QString str;
	str = index;

	p.setFont(font);

	p.drawPixmap( cx-7, cy-8, base );

	str.sprintf("%d", index);
	p.drawText(cx-2, cy+4, str);
}

bool Marker::isClip(QPixmap *scr, int sx, int sy)
{
	if(sy < M_TOP || sy > scr->height()-M_BOTTOM ||
	   sx < M_LEFT || sx > scr->width()-M_RIGHT)
		return true;

	return false;
}

void Marker::drawClipMarker(QPixmap *scr, int sx, int sy, int index)
{
	QPainter p(scr);
	int dx,dy;
	int mx,my;
	int ax,ay;
	int top, bottom, left, right;
	int arrow;

	int offsety[] = {
		-M_ARROW_LEN45, -M_ARROW_LEN, -M_ARROW_LEN45,
		0, 0, 0,
		M_ARROW_LEN45, M_ARROW_LEN, M_ARROW_LEN45,
	};
	int offsetx[] = {
		-M_ARROW_LEN45, 0, M_ARROW_LEN45,
		-M_ARROW_LEN, 0, M_ARROW_LEN,
		-M_ARROW_LEN45, 0, M_ARROW_LEN45,
	};

	mx = my = ax = ay = 0;

	top = M_TOP + M_ARROW_LEN;
	bottom = scr->height() - M_BOTTOM - M_ARROW_LEN;
	left = M_LEFT + M_ARROW_LEN;
	right = scr->width() - M_RIGHT - M_ARROW_LEN;

	mx = ax = sx; my = ay = sy;
	arrow = 4;

	if(sy < top){
		my = top;
		arrow = 1;
	}else if(sy > bottom){
		my = bottom;
		arrow = 7;
	}

	if(sx < left){
		mx = left;
		arrow--;
	}else if(sx > right){
		mx = right;
		arrow++;
	}

	ay = my + offsety[arrow];
	ax = mx + offsetx[arrow];

	p.drawPixmap(ax-3, ay-3, this->arrow[arrow]);
	p.end();

	drawBase(scr, mx, my, index);
	
}

void Marker::getMarkerpos(int dx, int dy, int max_x, int max_y,
						  int *mx, int *my, int *ax, int *ay)
{
	int x, y;

	if(dx == 0){
		*ax = *mx = 0;
		*ay = max_y;
		*my = *ay - M_ARROW_LEN;
		return;
	}
	if(dy == 0){
		*ay = *my = 0;
		*ax = max_x;
		*mx = *ax - M_ARROW_LEN;
	}

	y = dy*max_x/dx;
	if(y < max_y){
		*ay = y;
		*ax = max_x;
		*mx = max_x-M_ARROW_LEN;
		*my = dy*(*mx)/dx;
		return;
	}

	x = dx*max_y/dy;
	*ax = x;
	*ay = max_y;
	*my = max_y-M_ARROW_LEN;
	*mx = dx*(*my)/dy;
}

bool Marker::keyPressEvent ( QKeyEvent * e )
{
	switch(e->key()){
	case Qt::Key_M:
		marker_on = !marker_on;
		if(marker_on && !marker_bit){
			marker_bit=(1<<10)-1;
//			MARKER_TOGGLE(1);
		}
		parent->repaint(FALSE);
		dispMarkerInfo();
		return true;

	case Qt::Key_Q:
		if(marker_timer_id > 0){
			parent->killTimer(marker_timer_id);
			marker_timer_id=-1;
		}
		return true;
	case Qt::Key_0:
	case Qt::Key_1:
	case Qt::Key_2:
	case Qt::Key_3:
	case Qt::Key_4:
	case Qt::Key_5:
	case Qt::Key_6:
	case Qt::Key_7:
	case Qt::Key_8:
	case Qt::Key_9:
		if(marker_timer_id < 0)
			break;
		marker_on = true;

		MARKER_TOGGLE(e->key() - Qt::Key_0);
		if(!marker_bit)
			marker_on = false;
		parent->repaint(FALSE);
		dispMarkerInfo();
		return true;
	}
	return false;
}

bool Marker::timerEvent( QTimerEvent *e )
{
	int id;

	id = e->timerId();
	if(id == marker_timer_id){
		parent->killTimer(id);
		marker_timer_id=-1;
		return true;
	}
	return false;
}

void Marker::dispMarkerInfo()
{
	QString str;
	int i;

	if(marker_timer_id > 0){
		parent->killTimer(marker_timer_id);
		marker_timer_id=-1;
	}

	if(marker_on){
		str = "MarkerON:";
		for(i=0;i<10;i++){
			str.sprintf("%s<font color=\"%s\">%d</font>",
						str.ascii(), IS_MARKER_BIT(i)?"#000":"#999", i);
		}
	}else{
		str = "MarkerOFF";
	}

	((GMap *)parent)->dispInfo(str);
	marker_timer_id = parent->startTimer(3000);
}
