/*
  pos.cpp
  Copyright (C) 2008 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>
#include <qt.h>
#include <qregexp.h>
#include <qstring.h>
#include "pos.h"
#include "qgmap.h"

Pos::Pos(double _lat, double _lon, int _zoom)
{
	setZoom(_zoom);
	setLatLon(_lat, _lon);
}

Pos::Pos(int _x, int _y, int _zoom)
{
	setZoom(_zoom);
	setPos(_x, _y);
}

void Pos::setPos(int _x, int _y)
{
	x = _x;
	y = _y;
}

void Pos::setZoom(int _zoom)
{
	zoom = _zoom;
}

void Pos::setLatLon(double lat, double lon)
{
	int max_len = IMAGE_SIZE << (19-zoom);
	double dx, dy;

	while (lon < -180.0 || lon >= 180.0)
		lon = lon + ((lon<0.0)?360.0:-360.0);

	while (lat < -90.0 || lat >= 90.0)
		lat = lat + ((lat<0.0)?180.0:-180.0);

	dx = (lon+180.0)/360.0*(double)max_len;
	dy = (0.5-logf(tanf((lat+90.0)/180.0*3.141592/2.0))/(2.0*3.141592))
		*(double)max_len;

	x = (int)dx;
	y = (int)dy;
}

Pos &Pos::offset(int _x, int _y)
{
	static Pos retpos;

	retpos.setZoom(zoom);
	retpos.setPos(this->getX() + _x, this->getY() + _y);
  
//  printf("x=%d, y=%d\n", _x, _y);
	return retpos;
}

Pos &Pos::offset_clip(int _x, int _y)
{
	static Pos retpos;
	int max;

//	printf("Pos::offset_clip(%d,%d) x,y=%d,%d\n", _x, _y, getX(), getY());
	max = (1<<(19-zoom))*IMAGE_SIZE;
	_x += this->getX();
	_y += this->getY();
	if(_x < 0) _x = 0;
	if(_y < 0) _y = 0;
	if(_x >= max) _x = max-1;
	if(_y >= max) _y = max-1;
//	printf("Pos::offset_clip zoom=%d max=%d _x=%d, _y=%d\n", zoom, max, _x, _y);

	retpos.setZoom(zoom);
	retpos.setPos(_x, _y);

//	printf("x=%d, y=%d\n", _x, _y);
	return retpos;
}

double Pos::getLat()
{
	int zm = 1 << (19-zoom);
	double lat;

	/*
	  logf(tanf((lat+90.0)/180.0*3.141592/2.0))/(2.0*3.141592)) = 0.5 - dy/(256.0*(double)zm);
	  pow(e, 0.5 - dy/(256.0*(double)zm))*(2.0*3.141592) = tanf((lat+90.0)/180.0*3.141592/2.0));
	  (lat+90.0)/180.0*3.141592/2.0 = atan(pow(e, 0.5 - dy/(256.0*(double)zm))*(2.0*3.141592));
	  lat  = atan(pow(e, 0.5 - dy/(256.0*(double)zm))*(2.0*3.141592))*180.0/3.141592*2.0) - 90.0;
	*/

	lat = 90.0 - atan(pow(0.367879, (0.5-(double)y/((double)zm*256.0))*2.0*3.141592))
		/3.141592*180.0*2.0;

	return lat;
}

double Pos::getLon()
{
	int max_len = IMAGE_SIZE << (19-zoom);
	double lon;
	int _x = x;

	while(_x < 0)
		x+=max_len;
	x %= max_len;

	lon = (double)x/(double)max_len*360.0-180.0;

	return lon;
}


int Pos::getX() const
{
	return x;
}

int Pos::getY() const
{
	return y;
}

int Pos::getZoom() const
{
	return zoom;
}

bool Pos::zoomup(int _zoom)
{
	int new_zoom;
	int d;

	new_zoom = zoom + _zoom;
	if(new_zoom > 19) new_zoom = 19;
	if(new_zoom < 0) new_zoom = 0;

	if(new_zoom == zoom)
		return FALSE;

//	printf("_zoom = %d, zoom=%d,x=%d, y=%d\n", _zoom, zoom, x, y);
	if(_zoom < 0){
		// zoom in
		d = 1 << (-_zoom);
		x *= d;
		y *= d;
	}else{
		d = 1 << _zoom;
		x /= d;
		y /= d;
	}
	zoom = new_zoom;

//	printf("Pos::zoomup zoom=%d x,y=%d,%d\n", zoom, getX(), getY());

	return TRUE;
}

bool Pos::zoomTo(int _zoom)
{
	if(_zoom != zoom)
		return zoomup(_zoom - zoom);
	return TRUE;
}

void Pos::normalize()
{
	int max;

	max = (1<<(19-zoom))*IMAGE_SIZE;

	while(x<0)
		x+=max;
	x%=max;
}

#define EARTH_B2PA2								\
	((double)EARTH_RB/(double)EARTH_RA)			\
	* ((double)EARTH_RB/(double)EARTH_RA)

#define EARTH_2 ((1.0-EARTH_B2PA2)/2.0)
#define EARTH_1 (1.0-EARTH_2)

double Pos::getLatLength()
{
	double lat, latlen;
	double rad;

	lat = getLat();

	rad = lat * PI / 180.0;

	latlen = 2.0*PI*((double)EARTH_RA)*cos(rad)
		/sqrt(EARTH_1+EARTH_2*cos(2.0*rad));

//	printf("lat:%lf latlen:%lf\n", lat, latlen);

	return latlen;
}

double Pos::getDotLength()
{
	unsigned long int mapsize = IMAGE_SIZE << (19-zoom);
	double latlen=getLatLength();

	latlen /= (double)mapsize;

//	printf("dot:%lf\n", latlen);

	return latlen;
}

int Pos::getScaleSize(int max, int *mul)
{
	double dotlen = getDotLength();
	int maxlen = (int)(dotlen * max);

	*mul=1;
	while(maxlen >= 10){
		*mul *= 10;
		maxlen /= 10;
	}

	if(maxlen>=5)
		*mul *= 5;
	else if(maxlen >= 2)
		*mul *= 2;

	return (int)((double)*mul/dotlen);
}

bool Pos::setLatLon(QString str)
{
	QString strlat, strlon;
	int i;
	double lat,lon;

	i = str.find(',');
	if(i < 0) return false;

	strlat = str.left(i).stripWhiteSpace();
	strlon = str.mid(i+1).stripWhiteSpace();

	if(!parseStr(strlat, &lat))
		return false;
	if(!parseStr(strlon, &lon))
		return false;

	setLatLon(lat, lon);

	return true;
}

bool Pos::parseStr(const QString str, double *value)
{
	QRegExp dms("^[+-]?\\d?\\d?\\d\\.[0-5]?\\d\\.[0-5]?\\d");
	QRegExp deg("^[+-]?\\d?\\d?\\d\\.?\\d*$");
	bool ok;

//	printf("dms.match(%s) = %d\n", str.ascii(), dms.match(str));
//	printf("deg.match(%s) = %d\n", str.ascii(), deg.match(str));

	if(deg.match(str)==0){

		*value = str.toDouble(&ok);
		if(!ok) return false;

	}else if(dms.match(str)==0){
		QString d,m,s,work;
		int i;
		i = str.find('.');
		d = str.left(i);
		work = str.mid(i+1);

		i = work.find('.');
		m = work.left(i);
		s = work.mid(i+1);

//		printf("d:%s m:%s s:%s\n", d.ascii(), m.ascii(), s.ascii());

		*value = d.toDouble(&ok);
		if(!ok) return false;
		*value += m.toDouble(&ok)/60.0;
		if(!ok) return false;
		*value += m.toDouble(&ok)/3600.0;
		if(!ok) return false;

	}else{

		return false;

	}

//	printf("deg = %f\n", *value);
	return true;
}

QString Pos::serialize()
{
	QString str;

	str.sprintf("%d,%d,%d", x, y, zoom);

	return str;
}

bool Pos::deserialize(const QString str)
{
	QString sx,sy,szoom,work;
	int i;

	i = str.find(',');
	if(i < 0)
		return false;
	sx = str.left(i);
	work = str.mid(i+1);

	i = work.find(',');
	if(i < 0)
		return false;
	sy = work.left(i);
	szoom = work.mid(i+1);

	x = sx.toInt();
	y = sy.toInt();
	zoom = szoom.toInt();

	return true;
}
